local L = AceLibrary("AceLocale-2.2"):new("RaidBuffStatus")
local options = { 
	type='group',
	args = {
		show = {
			type = 'execute',
			name = 'show',
			desc = L["Show the buff report summary window."],
			usage = "show",
			func = function()
				RaidBuffStatus:DoReport()
				RaidBuffStatus:ShowReportFrame()
			end,
		},
		hide = {
			type = 'execute',
			name = 'hide',
			desc = L["Hide the buff report summary window."],
			usage = "hide",
			func = "HideReportFrame",
		},
		report = {
			type = 'execute',
			name = 'report',
			desc = L["Report to /raid or /party who is not buffed to the max."],
			usage = "report",
			func = function()
				RaidBuffStatus:DoReport()
				RaidBuffStatus:ReportToChat(true)
			end,
		},
    	},
}


RaidBuffStatus = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceDB-2.0","AceEvent-2.0","AceDebug-2.0")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local BS = AceLibrary("Babble-Spell-2.2")

local addon = RaidBuffStatus

addon:RegisterChatCommand("/rbs",options)
addon:RegisterChatCommand("/raidbuffstatus",options)
addon:RegisterDB("RaidBuffStatusDB")

--addon:SetDebugging(true)

-- The original lists of flasks and elixirs was taken from BigBrother

local flasks = {
	L["Chromatic Resistance"],
	L["Distilled Wisdom"],
	L["Flask of Fortification"],
	L["Flask of Mighty Restoration"],
	L["Flask of Relentless Assault"],
	L["Supreme Power"],
	L["Flask of the Titans"],
	L["Supreme Power of Shattrath"],
	L["Fortification of Shattrath"],
	L["Mighty Restoration of Shattrath"],
	L["Relentless Assault of Shattrath"],
	L["Flask of Pure Death"],
	L["Flask of Blinding Light"],
	L["Chromatic Wonder"],
	L["Unstable Flask of the Bandit"],
	L["Unstable Flask of the Beast"],
	L["Unstable Flask of the Elder"],
	L["Unstable Flask of the Physician"],
	L["Unstable Flask of the Soldier"],
	L["Unstable Flask of the Sorcerer"],
}

local belixirs = {
	L["Fel Strength Elixir"],
	L["Onslaught Elixir"],
	L["Major Strength"],
	L["Elixir of Major Strength"],
	L["Major Agility"],
	L["Elixir of Major Agility"],
	L["Elixir of Mastery"],
	L["Major Firepower"],
	L["Elixir of Major Firepower"],
	L["Major Shadow Power"],
	L["Elixir of Major Shadow Power"],
	L["Major Frost Power"],
	L["Elixir of Major Frost Power"],
	L["Adept's Elixir"],
	L["Arcane Elixir"],
	L["Greater Arcane Elixir"],
	L["Healing Power"],
	L["Elixir of Healing Power"],
	L["Elixir of the Mongoose"],
	L["Elixir of Greater Firepower"],
}
local gelixirs = {
	L["Major Armor"],
	L["Elixir of Major Defense"],
	L["Greater Armor"],
	L["Elixir of Superior Defense"],
	L["Greater Mana Regeneration"],
	L["Elixir of Major Mageblood"],
	L["Mana Regeneration"],
	L["Mageblood Potion"],
	L["Earthen Elixir"],
	L["Elixir of Draenic Wisdom"],
	L["Elixir of Major Fortitude"],
	L["Greater Intellect"],
	L["Elixir of Empowerment"],
	L["Empowerment"],
	L["Elixir of Ironskin"],
}

local foods = {
	L["Well Fed"],
	L["\"Well Fed\""],
	L["Enlightened"],
}

local fortitude = {
	BS["Power Word: Fortitude"],
	BS["Prayer of Fortitude"],
}

local wild = {
	BS["Mark of the Wild"],
	BS["Gift of the Wild"],
}

local intellect = {
	BS["Arcane Intellect"],
	BS["Arcane Brilliance"],
}

local spirit = {
	BS["Divine Spirit"],
	BS["Prayer of Spirit"],
}

local shadow = {
	BS["Shadow Protection"],
	BS["Prayer of Shadow Protection"],
}

local auras = {
	BS["Crusader Aura"],
	BS["Devotion Aura"],
	BS["Retribution Aura"],
	BS["Concentration Aura"],
	BS["Sanctity Aura"],
	BS["Shadow Resistance Aura"],
	BS["Frost Resistance Aura"],
	BS["Fire Resistance Aura"],
}

local classes = {
	"PRIEST",
	"DRUID",
	"PALADIN",
	"ROGUE",
	"MAGE",
	"WARLOCK",
	"SHAMAN",
	"WARRIOR",
	"HUNTER",
}


local raid = {
	classes = {
		PRIEST = {},
		DRUID = {},
		PALADIN = {},
		ROGUE = {},
		MAGE = {},
		WARLOCK = {},
		SHAMAN = {},
		WARRIOR = {},
		HUNTER = {},
	},
	SpiritTalent = {},
	ClassNumbers = {},
	israid = false,
	isparty = false,
	readid = 0,
}
raid.reset = function()
	raid.classes = {}
	raid.classes.PRIEST = {}
	raid.classes.DRUID = {}
	raid.classes.PALADIN = {}
	raid.classes.ROGUE = {}
	raid.classes.MAGE = {}
	raid.classes.WARLOCK = {}
	raid.classes.SHAMAN = {}
	raid.classes.WARRIOR = {}
	raid.classes.HUNTER = {}
	raid.SpiritTalent = {}
	raid.ClassNumbers = {}
	raid.israid = false
	raid.isparty = false
	raid.readid = 0
	addon:Debug("raid reset")
end

local talents = {}

local report = {
	FoodList = {},
	IntellectList = {},
	WildList = {},
	SpiritList = {},
	FortitudeList = {},
	ShadowList = {},
	FlaskList = {},
	BElixirList = {},
	GElixirList = {},
	CrusaderList = {},
	PVPList = {},
	OfflineList = {},
	AFKList = {},
	ShadowsList = {},
	NoAuraAtAllList = {},
	GroupWithTooFewAuraList = {},
	DeadList = {},
	ZoneList = {},
	HealthList = {},
	ManaList = {},
	someoneneedsbuff = 0,
	warnings = 0,
	checking = {},
}

report.reset = function()
	report.FoodList = {}
	report.IntellectList = {}
	report.WildList = {}
	report.SpiritList = {}
	report.FortitudeList = {}
	report.ShadowList = {}
	report.FlaskList = {}	
	report.BElixirList = {}
	report.GElixirList = {}
	report.CrusaderList = {}
	report.PVPList = {}
	report.OfflineList = {}
	report.AFKList = {}
	report.ShadowsList = {}
	report.NoAuraAtAllList = {}
	report.GroupWithTooFewAuraList = {}
	report.DeadList = {}
	report.ZoneList = {}
	report.HealthList = {}
	report.ManaList = {}
	report.someoneneedsbuff = 0
	report.warnings = 0
	report.checking = {}
	addon:Debug("report reset")
end
report.morewarnings = function()
	report.warnings = report.warnings + 1
end
report.morebuffs = function()
	report.someoneneedsbuff = report.someoneneedsbuff + 1
end

local PriestAnnounced = {}
local inspectqueuename = ""
local inspectqueueclass = ""
local inspectqueuetime = 0
local buttons = {}

addon:RegisterDefaults("profile", {
	CheckFood = true,
	CheckIntellect = true,
	CheckWild = true,
	CheckSpirit = true,
	CheckFortitude = true,
	CheckShadow = true,
	CheckFlaskir = true,
	CheckCrusader = true,
	CheckPVP = true,
	CheckOffline = true,
	CheckAFK = true,
	CheckShadows = true,
	CheckAuras = true,
	CheckNoAura = true,
	CheckDead = true,
	CheckZone = true,
	CheckHealth = true,
	CheckMana = true,
	CheckAuto = true,
	ReportSelf = false,
	ReportChat = true,
	LockWindow = false,
	x = 0,
	y = 0,
})


-- End of inits


function addon:ShowReportFrame()
	self:Debug("Show.");
	ShowUIPanel(RBSFrame);
end

function addon:HideReportFrame()
	addon:Debug("Hide.");
	HideUIPanel(RBSFrame);
end


function addon:DoReport()
	addon:Debug("Report starting.")
	report:reset()
	addon:Debug("Reading raid")
	addon:ReadRaid()
	if (not raid.israid) and (not raid.isparty) then
		addon:Debug("Report aborted as not in raid or party.")
		return
	end
	addon:Debug("Calculating report")
	addon:CalculateReport()
	addon:Debug("Updating buttons")
	addon:UpdateButtons()
end

function addon:CalculateReport()
	local thiszone = GetRealZoneText()
	local PalaPerGroupList = { {}, {}, {}, {}, {}, {}, {}, {} }
	local PalaAuraPerGroupList = { {}, {}, {}, {}, {}, {}, {}, {} }
	for class,_ in pairs(raid.classes) do
		addon:Debug("Class:" .. class .. " " .. raid.ClassNumbers[class])
		for name,_ in pairs(raid.classes[class]) do
			addon:Debug("Name:" .. name .. " ...........")
			local unit = raid.classes[class][name]
			local missingbuff = false

			if not unit.online then
				if addon.db.profile.CheckOffline then
					table.insert(report.OfflineList, name)
					report.morewarnings()
					addon:Debug("offline")
				end
				break
			end

			if addon.db.profile.CheckPVP then
				if UnitIsPVP(unit.unitid) then
					table.insert(report.PVPList, name)
					report.morewarnings()
					addon:Debug("pvp")
				end
			end

			if addon.db.profile.CheckAFK then
				if UnitIsAFK(unit.unitid) then
					table.insert(report.AFKList, name)
					report.morewarnings()
					addon:Debug("afk")
				end
			end
		
			if addon.db.profile.CheckZone and raid.israid then
				if thiszone ~= unit.zone then
					table.insert(report.ZoneList, name)
					report.morewarnings()
					addon:Debug("zone")
				end
			end

			if addon.db.profile.CheckDead then
				if unit.isdead then
					table.insert(report.DeadList, name)
					report.morewarnings()
					addon:Debug("dead")
				end
			end

			if addon.db.profile.CheckHealth and not unit.isdead then
				local maxhealth = UnitHealthMax(unit.unitid)
				local health = UnitHealth(unit.unitid)
				local percent = (health/maxhealth)*100
				if percent < 80 then
					table.insert(report.HealthList, name)
					report.morewarnings()
					addon:Debug("health")
				end
			end

			if addon.db.profile.CheckMana and not unit.isdead then
				if class ~= "WARRIOR" and class ~= "ROGUE" then
					local maxmana = UnitManaMax(unit.unitid)
					local mana = UnitMana(unit.unitid)
					local percent = (mana/maxmana)*100
					if percent < 80 then
						table.insert(report.ManaList, name)
						report.morewarnings()
						addon:Debug("mana")
					end
				end
			end

			if addon.db.profile.CheckCrusader then
				if class == "PALADIN" then
					report.checking.crusader = true
					if Aura:UnitHasBuff(unit.unitid, BS["Crusader Aura"]) then
						table.insert(report.CrusaderList, name)
						report.morewarnings()
						addon:Debug("crusader")
					end
				end
			end

			if addon.db.profile.CheckShadows then
				if # raid.classes.PRIEST > 0 then
					if class == "PALADIN" then
						report.checking.shadows = true
						if Aura:UnitHasBuff(unit.unitid, BS["Shadow Resistance Aura"]) then
							local hasbuff = false
							for _, v in ipairs(shadow) do
								if Aura:UnitHasBuff(unit.unitid, v) then
									hasbuff = true
									break
								end
							end
							if hasbuff then
								table.insert(report.ShadowsList, name)
								report.morewarnings()
								addon:Debug("2 shadow")
							end
						end
					end
				end
			end

			if addon.db.profile.CheckAuras or addon.db.profile.CheckNoAura then
				if class == "PALADIN" then
					report.checking.auras = true
					table.insert(PalaPerGroupList[unit.group], name)
					local palahasaura = false
					for _, v in ipairs(auras) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							palahasaura = true
							local foundauraingroup = false
							for _, k in ipairs(PalaAuraPerGroupList[unit.group]) do
								if v == k then
									foundauraingroup = true
									break
								end
							end
							if not foundauraingroup then
								table.insert(PalaAuraPerGroupList[unit.group], v)
							end
						end
					end
					if addon.db.profile.CheckNoAura and (not palahasaura) then
						table.insert(report.NoAuraAtAllList, name)
						report.morebuffs()
						addon:Debug("no aura")
					end
				end
			end

			if addon.db.profile.CheckFlaskir then
				missingbuff = true
				for _, v in ipairs(flasks) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					local numbbelixir = 0
					local numbgelixir = 0
					for _, v in ipairs(belixirs) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							numbbelixir = 1
							break
						end
					end
					for _, v in ipairs(gelixirs) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							numbgelixir = 1
							break
						end
					end
					local totalelixir = numbbelixir + numbgelixir
					if totalelixir == 0 then
						table.insert(report.FlaskList, name) -- no flask or elixir
						report.morebuffs()
						addon:Debug("flask/elix")
					elseif totalelixir == 1 then
						report.morebuffs()
						addon:Debug("elix")
						if numbbelixir == 0 then
							table.insert(report.BElixirList, name)
							addon:Debug("belix")
						else
							table.insert(report.GElixirList, name)
							addon:Debug("gelix")
						end
					end
				end
			end

			if addon.db.profile.CheckFood then
				missingbuff = true
				for _, v in ipairs(foods) do
					if Aura:UnitHasBuff(unit.unitid, v) then
						missingbuff = false
						break
					end
				end
				if missingbuff then
					table.insert(report.FoodList, name)
					report.morebuffs()
					addon:Debug("food")
				end
			end
		
			if addon.db.profile.CheckIntellect then
				if not addon.db.profile.CheckAuto or (addon.db.profile.CheckAuto and raid.ClassNumbers.MAGE > 0) then
					report.checking.intellect = true
					missingbuff = true
					for _, v in ipairs(intellect) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						table.insert(report.IntellectList, name)
						report.morebuffs()
						addon:Debug("int")
					end
				end
			end
	
			if addon.db.profile.CheckWild then
				if not addon.db.profile.CheckAuto or (addon.db.profile.CheckAuto and raid.ClassNumbers.DRUID > 0) then
					report.checking.wild = true
					missingbuff = true
					for _, v in ipairs(wild) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						table.insert(report.WildList, name)
						report.morebuffs()
						addon:Debug("wild")
					end
				end
			end


			if addon.db.profile.CheckSpirit then
				if not addon.db.profile.CheckAuto or (addon.db.profile.CheckAuto and # raid.SpiritTalent > 0) then
					report.checking.spirit = true
					missingbuff = true
					for _, v in ipairs(spirit) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						table.insert(report.SpiritList, name)
						report.morebuffs()
						addon:Debug("spirit")
					end
				end
			end

			if addon.db.profile.CheckFortitude then
				if not addon.db.profile.CheckAuto or (addon.db.profile.CheckAuto and raid.ClassNumbers.PRIEST > 0) then
					report.checking.fortitude = true
					missingbuff = true
					for _, v in ipairs(fortitude) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						table.insert(report.FortitudeList, name)
						report.morebuffs()
						addon:Debug("fortitude")
					end
				end
			end
		
			if addon.db.profile.CheckShadow then
				if not addon.db.profile.CheckAuto or (addon.db.profile.CheckAuto and raid.ClassNumbers.PRIEST > 0) then
					report.checking.shadow = true
					missingbuff = true
					for _, v in ipairs(shadow) do
						if Aura:UnitHasBuff(unit.unitid, v) then
							missingbuff = false
							break
						end
					end
					if missingbuff then
						table.insert(report.ShadowList, name)
						report.morebuffs()
						addon:Debug("shadow")
					end
				end
			end
		end
	end

	if addon.db.profile.CheckAuras then
		for i, v in ipairs(PalaPerGroupList) do
			if # v > 1 then
				if # v > # PalaAuraPerGroupList[i] then
					table.insert(report.GroupWithTooFewAuraList, i)
					report.morewarnings()
					addon:Debug("not enough aura:" .. i)
				end
			end
		end
	end
end

function addon:ReportToChat(boss)
	local needsbuff = report.someoneneedsbuff
	if boss then
		addon:Debug("Boss reporting")
	else
		addon:Debug("Trash reporting")
		needsbuff = needsbuff - # report.FlaskList - # report.GElixirList - # report.BElixirList
	end

	if report.warnings > 0 then
		addon:Say(L["Warnings: "] .. report.warnings)
		if # report.CrusaderList > 0 then
			addon:Say(L["<Crusader Aura>: "] .. table.concat(report.CrusaderList, ", "))
		end
		if # report.PVPList > 0 then
			addon:Say(L["<PVP On>: "] .. table.concat(report.PVPList, ", "))
		end
		if # report.AFKList > 0 then
			addon:Say(L["<AFK>: "] .. table.concat(report.AFKList, ", "))
		end
		if # report.OfflineList > 0 then
			addon:Say(L["<Offline>: "] .. table.concat(report.OfflineList, ", "))
		end
		if # report.ShadowsList > 0 then
			addon:Say(L["<Shadow Resistance Aura AND Shadow Protection>: "] .. table.concat(report.ShadowsList, ", "))
		end
		if # report.DeadList > 0 then
			addon:Say(L["<Dead>: "] .. table.concat(report.DeadList, ", "))
		end
		if # report.ZoneList > 0 then
			addon:Say(L["<Different Zone>: "] .. table.concat(report.ZoneList, ", "))
		end
		if # report.HealthList > 0 then
			addon:Say(L["<Health less than 80%%>: "] .. table.concat(report.HealthList, ", "))
		end
		if # report.ManaList > 0 then
			addon:Say(L["<Mana less than 80%%>: "] .. table.concat(report.ManaList, ", "))
		end
		if # report.GroupWithTooFewAuraList > 0 then
			for _, v in ipairs(report.GroupWithTooFewAuraList) do
				local palalist = {}
				for i, k in pairs(raid.classes.PALADIN) do
					if k.group == v then
						table.insert(palalist, i)
					end
				end
				addon:Say(L["<Paladin Different Aura>: Group "] .. v .. ": " .. table.concat(palalist, ", "))
			end
		end

	end
	if needsbuff > 0 then
		if boss then
			addon:Say(L["Missing buffs (Boss): "] .. needsbuff)
		else
			addon:Say(L["Missing buffs (Trash): "] .. needsbuff)
		end
		if # report.FoodList > 0 then
			addon:Say(L["<Well Fed>: "] .. table.concat(report.FoodList, ", "))
		end
		if # report.IntellectList > 0 then
			addon:Say(L["<Arcane Intellect>: "] .. table.concat(report.IntellectList, ", "))
		end
		if # report.WildList > 0 then
			addon:Say(L["<Mark of the Wild>: "] .. table.concat(report.WildList, ", "))
		end
		if # report.SpiritList > 0 then
			addon:Say(L["<Divine Spirit>: "] .. table.concat(report.SpiritList, ", "))
		end
		if # report.FortitudeList > 0 then
			addon:Say(L["<Power Word: Fortitude>: "] .. table.concat(report.FortitudeList, ", "))
		end
		if # report.ShadowList > 0 then
			addon:Say(L["<Shadow Protection>: "] .. table.concat(report.ShadowList, ", "))
		end
		if boss then
			if # report.FlaskList > 0 then
				addon:Say(L["<Flask or two Elixirs>: "] .. table.concat(report.FlaskList, ", "))
			end
			if # report.BElixirList > 0 then
				addon:Say(L["<Battle Elixir>: "] .. table.concat(report.BElixirList, ", "))
			end
			if # report.GElixirList > 0 then
				addon:Say(L["<Guardian Elixir>: "] .. table.concat(report.GElixirList, ", "))
			end
		end
		if # report.NoAuraAtAllList > 0 then
			addon:Say(L["<Paladin Aura>: "] .. table.concat(report.NoAuraAtAllList, ", "))
		end
	else
		addon:Say(L["No buffs needed!"])
	end
end




function addon:ReadRaid()
	raid.readid = raid.readid + 1
	local raidnum = GetNumRaidMembers()
	local partynum = GetNumPartyMembers()
	if raidnum < 2 then
		if partynum < 1 then
			raid.reset()
			return
		else
			addon:Debug("Party detected.")
			raid.isparty = true
			raid.israid = false
			for i = 1, partynum do
				addon:ReadUnit("party" .. i)
			end
			addon:ReadUnit("player")
		end
	else
		if raid.isparty then
			addon:Debug("Party has converted to Raid!")
			raid.reset()
		else
			addon:Debug("Raid detected.")
		end
		raid.isparty = false
		raid.israid = true
		for i = 1, raidnum do
			addon:ReadUnit("raid" .. i)
		end
	end
	addon:DeleteOldUnits()
	addon:CalculateClassNumbers()
	if addon.db.profile.CheckAuto then
		addon:FindPriestsWithSpirit()
	end
end


-- raid = { classes = { CLASS = { NAME = { readid, unitid, group, zone, online, isdead, talents = {}, 
function addon:ReadUnit(unitid)
	if not UnitExists(unitid) then
		addon:Debug("Unit does not exist:" .. unitid)
		return
	end
	local unitindex = select(3, unitid:find("(%d+)"))
	local name = UnitName(unitid)
	if name and name ~= UNKNOWNOBJECT and name ~= UKNOWNBEING then

		local class = select(2, UnitClass(unitid))
		local isDead = UnitIsDeadOrGhost(unitid) or false
		local rank = 0
		local subgroup = 1
		local online =  UnitIsConnected(unitid)
		local role = ""
		local zone = "UNKNOWN"
		local nametwo = name
		local isML = false
		local level = UnitLevel(unitid)
		
		if raid.israid then
			nametwo, rank, subgroup, _, _, _, zone, _, _, role, isML = GetRaidRosterInfo(unitindex)
			if name ~= nametwo then
				addon:Debug("NAMES DIFFER!" .. name .. " " .. nametwo)
			end
		end
		if raid.classes[class][name] == nil then
			raid.classes[class][name] = {}
		end
		local rcn = raid.classes[class][name]
		rcn.readid = raid.readid
		rcn.unitid = unitid
		rcn.group = subgroup
		rcn.zone = zone
		rcn.online = online
		rcn.isdead = isDead
		rcn.role = role
		rcn.rank = rank
		rcn.level = level
	end
end

function addon:DeleteOldUnits()
	addon:Debug("Delete old units.")
	for class,_ in pairs(raid.classes) do
		for name,_ in pairs(raid.classes[class]) do
			if raid.classes[class][name].readid < raid.readid then
				addon:Debug("Left raid:" .. name)
				raid.classes[class][name] = nil
			end
		end
	end
end

function addon:FindPriestsWithSpirit()
	raid.SpiritTalent = {}
	for name,_ in pairs(raid.classes.PRIEST) do
		if raid.classes.PRIEST[name].talents == nil then
			if UnitIsVisible(name) and inspectqueuetime < GetTime() then
				addon:Debug("Inspect request going to be sent for: " .. name);
				inspectqueuetime = GetTime() + 15
				inspectqueuename = name
				inspectqueueclass = "PRIEST"
				addon:RegisterEvent("INSPECT_TALENT_READY", "ReadTalents")
				NotifyInspect(name)
			end
		else
			addon:Debug("Inspect was cached: " .. name);
			local spiritrank = raid.classes.PRIEST[name].talents.tree[1].talent[14]
			if spiritrank == nil then
				addon:Debug("talent 1 14 nil " .. name);  -- should never happen
			else
				if spiritrank > 0 then
					if PriestAnnounced[name] == nil then
						addon:Say(L["Priest with Spirit talent has been detected in the raid: "] .. name)
						PriestAnnounced[name] = true
					end
					table.insert(raid.SpiritTalent, name)
				else
					addon:Debug("No spirit talent found: " .. name)
				end
			end
		end
	end
end



function addon:ReadTalents()
	addon:Debug("Inspect request recieved: " .. inspectqueuename);
	self:UnregisterEvent("INSPECT_TALENT_READY")
	raid.classes.PRIEST[inspectqueuename].talents = { tree = {} }
	local tree = raid.classes.PRIEST[inspectqueuename].talents.tree
	for tab = 1, GetNumTalentTabs(true) do
		local name, icon, spent = GetTalentTabInfo(tab, true);
		tree[tab] = {name=name, icon=icon, spent=spent, talent={ }};
		for talent = 1, GetNumTalents(tab, true) do
			tree[tab].talent[talent] = select(5, GetTalentInfo(tab, talent, true));
--			addon:Debug(tab .. " " .. talent .. " " .. tree[tab].talent[talent])
		end
	end
	inspectqueuetime = 0
end


function addon:CalculateClassNumbers()
	for _,class in ipairs(classes) do
		raid.ClassNumbers[class] = addon:DicSize(raid.classes[class])
	end
end


function addon:Say(fmt, ...)
	local msg = string.format(fmt, ...)
	local str = "RBS::"
	local canspeak = IsRaidLeader() or IsRaidOfficer()
	for _,s in pairs({strsplit(" ", msg)}) do
		if #str + #s >= 250 then
			if addon.db.profile.ReportChat and raid.isparty then SendChatMessage(str, "party") end
			if addon.db.profile.ReportChat and raid.israid and canspeak then SendChatMessage(str, "raid") end
			if addon.db.profile.ReportSelf then addon:Print(str) end
			str = "RBS::"
		end
		str = str .. " " .. s
	end
	if addon.db.profile.ReportChat and raid.isparty then SendChatMessage(str, "party") end
	if addon.db.profile.ReportChat and raid.israid and canspeak then SendChatMessage(str, "raid") end
	if addon.db.profile.ReportSelf then addon:Print(str) end
end


function addon:DicSize(dic) -- is there really no built-in function to do this??
	local i = 0
	for _,_ in pairs(dic) do
		i = i + 1
	end
	return i
end


function addon:SetupFrames()
	addon.frame = CreateFrame("Frame", "RBSFrame", UIParent)
	addon.frame:Hide()
	addon.frame:EnableMouse(true)
	addon.frame:SetFrameStrata("MEDIUM")
	addon.frame:SetMovable(true)
	addon.frame:SetWidth(128)
	addon.frame:SetHeight(150)
	addon.rbsfs = addon.frame:CreateFontString("$parentTitle","ARTWORK","GameFontNormal")
	addon.rbsfs:SetText("RaidBuffStatus")
	addon.rbsfs:SetPoint("TOP",0,-5)
	addon.rbsfs:SetTextColor(200,0,0,150)
	addon.rbsfs:Show()
	addon.frame:SetBackdrop( { 
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	addon.frame:SetBackdropBorderColor(0,0,0,200)
	addon.frame:SetBackdropColor(0,0,0,100)
	addon.frame:ClearAllPoints()
	addon.frame:SetScript("OnMouseDown",function()
		if ( arg1 == "LeftButton" ) then
			if not addon.db.profile.LockWindow then
				this:StartMoving()
			end
		end
	end)
	addon.frame:SetScript("OnMouseUp",function()
		if ( arg1 == "LeftButton" ) then
			this:StopMovingOrSizing()
			addon:SaveFramePosition()
		end
	end)
	addon.frame:SetScript("OnHide",function() this:StopMovingOrSizing() end)
	addon.frame:SetClampedToScreen(true)
	addon:LoadFramePosition()
	addon:AddBuffButtons()
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["Boss"])
	addon.button:SetWidth(54)
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", 10, 5)
	addon.button:SetScript("OnClick", function() addon:DoReport() addon:ReportToChat(true) end)
	addon.button:Show()
	
	addon.button = CreateFrame("Button", "", addon.frame, "OptionsButtonTemplate")
	addon.button:SetText(L["Trash"])
	addon.button:SetWidth(54)
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", 64, 5)
	addon.button:SetScript("OnClick", function() addon:DoReport() addon:ReportToChat(false) end)
	addon.button:Show()
end


function addon:SaveFramePosition()
	addon.db.profile.x = addon.frame:GetLeft()
	addon.db.profile.y = addon.frame:GetTop() - UIParent:GetTop()
end

function addon:LoadFramePosition()
	addon.frame:ClearAllPoints()
	if (addon.db.profile.x ~= 0) or (addon.db.profile.y ~= 0) then
		addon.frame:SetPoint("TOPLEFT", UIParent,"TOPLEFT", addon.db.profile.x, addon.db.profile.y)
	else
		addon.frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	end
end


function addon:AddBuffButtons()
	addon:AddBuffButton(
		pvp,
		10, 102,
		"Interface\\Icons\\INV_BannerPVP_02",
		L["<PVP On>: "],
		function(self)
			if addon.db.profile.CheckPVP then
				self.count:SetText(# report.PVPList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckPVP")
		end,
		function(self)
			addon:Tooltip(self, L["PVP is On"], report.PVPList)
		end
	)
	addon:AddBuffButton(
		crusader,
		32, 102,
		BS:GetSpellIcon(BS["Crusader Aura"]),
		L["<Crusader Aura>: "],
		function(self)
			if addon.db.profile.CheckCrusader and report.checking.crusader then
				self.count:SetText(# report.CrusaderList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckCrusader")
		end,
		function(self)
			addon:Tooltip(self, L["Loladin has Crusader Aura"], report.CrusaderList)
		end
	)

	addon:AddBuffButton(
		shadows,
		54, 102,
		BS:GetSpellIcon(BS["Shadow Resistance Aura"]),
		L["<Shadow Resistance Aura AND Shadow Protection>: "],
		function(self)
			if addon.db.profile.CheckShadows and report.checking.shadows then
				self.count:SetText(# report.ShadowsList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckShadows")
		end,
		function(self)
			addon:Tooltip(self, L["Paladin has Shadow Resistance Aura AND Shadow Protection"], report.ShadowsList)
		end
	)
	addon:AddBuffButton(
		health,
		76, 102,
		"Interface\\Icons\\INV_Potion_131",
		L["<Health less than 80%%>: "],
		function(self)
			if addon.db.profile.CheckHealth then
				self.count:SetText(# report.HealthList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckHealth")
		end,
		function(self)
			addon:Tooltip(self, L["Player has health less than 80%"], report.HealthList)
		end
	)
	addon:AddBuffButton(
		mana,
		98, 102,
		"Interface\\Icons\\INV_Potion_137",
		L["<Mana less than 80%%>: "],
		function(self)
			if addon.db.profile.CheckMana then
				self.count:SetText(# report.ManaList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckMana")
		end,
		function(self)
			addon:Tooltip(self, L["Player has mana less than 80%"], report.ManaList)
		end
	)
	addon:AddBuffButton(
		zone,
		10, 80,
		"Interface\\Icons\\INV_Misc_QuestionMark",
		L["<Different Zone>: "],
		function(self)
			if addon.db.profile.CheckZone then
				self.count:SetText(# report.ZoneList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckZone")
		end,
		function(self)
			addon:Tooltip(self, L["Player is in a different zone"], report.ZoneList)
		end
	)
	addon:AddBuffButton(
		offline,
		32, 80,
		"Interface\\Icons\\INV_Gizmo_FelStabilizer",
		L["<Offline>: "],
		function(self)
			if addon.db.profile.CheckOffline then
				self.count:SetText(# report.OfflineList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckOffline")
		end,
		function(self)
			addon:Tooltip(self, L["Player is Offline"], report.OfflineList)
		end
	)
	addon:AddBuffButton(
		afk,
		54, 80,
		"Interface\\Icons\\Trade_Fishing",
		L["<AFK>: "],
		function(self)
			if addon.db.profile.CheckAFK then
				self.count:SetText(# report.AFKList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckAFK")
		end,
		function(self)
			addon:Tooltip(self, L["Player is AFK"], report.AFKList)
		end
	)
	addon:AddBuffButton(
		dead,
		76, 80,
		"Interface\\Icons\\Spell_Holy_SenseUndead",
		L["<Dead>: "],
		function(self)
			if addon.db.profile.CheckDead then
				self.count:SetText(# report.DeadList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckDead")
		end,
		function(self)
			addon:Tooltip(self, L["Player is Dead"], report.DeadList)
		end
	)
	addon:AddBuffButton(
		groupwithtoofewaura,
		98, 80,
		BS:GetSpellIcon(BS["Retribution Aura"]),
		L["<Paladin Different Aura>: Group "],
		function(self)
			if addon.db.profile.CheckAuras and report.checking.auras then
				self.count:SetText(# report.GroupWithTooFewAuraList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckAuras")
		end,
		function(self)
			addon:Tooltip(self, L["There are more Paladins than different Auras in group"], report.GroupWithTooFewAuraList)
		end
	)

	addon:AddBuffButton(
		food,					-- name
		10, 52,					-- x,y
		"Interface\\Icons\\INV_Misc_Food_74",	-- icon
		L["Well Fed"],				-- spell
		function(self)				-- update
			if addon.db.profile.CheckFood then
				self.count:SetText(# report.FoodList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)		-- click
			addon:ToggleCheck("CheckFood")
		end,
		function(self)				-- tooltip
			addon:Tooltip(self, L["Not Well Fed"], report.FoodList)
		end
	)

	addon:AddBuffButton(
		flask,
		32, 52,
		"Interface\\Icons\\INV_Potion_119",
		L["<Flask or two Elixirs>: "],
		function(self)
			if addon.db.profile.CheckFlaskir then
				self.count:SetText(# report.FlaskList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckFlaskir")
		end,
		function(self)
			addon:Tooltip(self, L["Missing a Flask or two Elixirs"], report.FlaskList)
		end
	)
	addon:AddBuffButton(
		belixir,
		54, 52,
		"Interface\\Icons\\INV_Potion_111",
		L["<Battle Elixir>: "],
		function(self)
			if addon.db.profile.CheckFlaskir then
				self.count:SetText(# report.BElixirList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckFlaskir")
		end,
		function(self)
			addon:Tooltip(self, L["Missing a Battle Elixir"], report.BElixirList)
		end
	)
	addon:AddBuffButton(
		gelixir,
		76, 52,
		"Interface\\Icons\\INV_Potion_158",
		L["<Guardian Elixir>: "],
		function(self)
			if addon.db.profile.CheckFlaskir then
				self.count:SetText(# report.GElixirList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckFlaskir")
		end,
		function(self)
			addon:Tooltip(self, L["Missing a Guardian Elixir"], report.GElixirList)
		end
	)
	addon:AddBuffButton(
		spirit,
		98, 52,
		nil,
		BS["Divine Spirit"],
		function(self)
			if addon.db.profile.CheckSpirit and report.checking.spirit then
				self.count:SetText(# report.SpiritList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckSpirit")
		end,
		function(self)
			addon:Tooltip(self, L["Missing Divine Spirit"], report.SpiritList)
		end
	)

	addon:AddBuffButton(
		intellect,
		10, 30,
		nil,
		BS["Arcane Intellect"],
		function(self)
			if addon.db.profile.CheckIntellect and report.checking.intellect then
				self.count:SetText(# report.IntellectList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckIntellect")
		end,
		function(self)
			addon:Tooltip(self, L["Missing Arcane Intellect"], report.IntellectList)
		end
	)
	addon:AddBuffButton(
		wild,
		32, 30,
		nil,
		BS["Mark of the Wild"],
		function(self)
			if addon.db.profile.CheckWild and report.checking.wild then
				self.count:SetText(# report.WildList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckWild")
		end,
		function(self)
			addon:Tooltip(self, L["Missing Mark of the Wild"], report.WildList)
		end
	)
	addon:AddBuffButton(
		fortitude,
		54, 30,
		nil,
		BS["Power Word: Fortitude"],
		function(self)
			if addon.db.profile.CheckFortitude and report.checking.fortitude then
				self.count:SetText(# report.FortitudeList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckFortitude")
		end,
		function(self)
			addon:Tooltip(self, L["Missing Power Word: Fortitude"], report.FortitudeList)
		end
	)
	addon:AddBuffButton(
		shadow,
		76, 30,
		nil,
		BS["Shadow Protection"],
		function(self)
			if addon.db.profile.CheckShadow and report.checking.shadow then
				self.count:SetText(# report.ShadowList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckShadow")
		end,
		function(self)
			addon:Tooltip(self, L["Missing Shadow Protection"], report.ShadowList)
		end
	)
	addon:AddBuffButton(
		noauraatall,				-- name
		98, 30,					-- x,y
		BS:GetSpellIcon(BS["Devotion Aura"]),	-- icon
		L["<Paladin Aura>: "],			-- spell
		function(self)
			if addon.db.profile.CheckNoAura and report.checking.auras then
				self.count:SetText(# report.NoAuraAtAllList)
			else
				self.count:SetText("")
			end
		end,
		function(self, button, down)
			addon:ToggleCheck("CheckNoAura")
		end,
		function(self)				-- tooltip
			addon:Tooltip(self, L["Paladin has no Aura at all"], report.NoAuraAtAllList)
		end
	)
end



function addon:AddBuffButton(name, x, y, icon, spell, update, click, tooltip)
	addon.button = CreateFrame("Button", name, addon.frame, "SecureActionButtonTemplate")
	table.insert(buttons, addon.button)
	addon.button:Hide()
	addon.button:SetWidth(20)
	addon.button:SetHeight(20)
	addon.button:SetNormalTexture(icon or BS:GetSpellIcon(spell) or "Interface\\Icons\\INV_ValentinesCandy")
	addon.count = addon.button:CreateFontString(nil, "ARTWORK","GameFontNormalSmall")
	addon.button.count = addon.count
	addon.count:SetWidth(20)
	addon.count:SetHeight(20)
	addon.count:SetFont(addon.count:GetFont(),12,"OUTLINE")
	addon.count:SetPoint("CENTER", self.button, "CENTER", 0, 0)
	addon.count:SetText("X")
	addon.count:Show()
	addon.button:SetPoint("BOTTOMLEFT", addon.frame, "BOTTOMLEFT", x, y)
	if click then
		addon.button:SetScript("OnClick", click)
	end
	if update then
		addon.button.update = update
	end
	if tooltip then
		addon.button:SetScript("OnEnter", tooltip)
		addon.button:SetScript("OnLeave", function() GameTooltip:Hide() end)
	end
	addon.button:Show()
end

function addon:ToggleCheck(...)
	if addon.db.profile[...] then
		addon.db.profile[...] = false
	else
		addon.db.profile[...] = true
		addon:DoReport()
	end
	addon:UpdateButtons()
end


function addon:UpdateButtons()
	for _,v in ipairs(buttons) do
		v:update(v)
	end
end

function addon:OnProfileEnable()
	addon:Debug("Profile changed")
	addon:LoadFramePosition()
	addon:DoReport()
end

function addon:JoinedPartyRaid()
	addon:Debug("JoinedPartyRaid");
	local oldstatus = raid.isparty or raid.israid
	addon:DoReport()
	if raid.isparty or raid.israid then
		addon:ShowReportFrame()
		addon:RegisterEvent("PLAYER_AURAS_CHANGED", "DoReport")
		addon:ScheduleRepeatingEvent("DoReport", addon.DoReport, 6, addon)
	elseif oldstatus then  -- was party or raid and now gone
		addon:HideReportFrame()
		addon:UnregisterEvent("PLAYER_AURAS_CHANGED")
		addon:CancelScheduledEvent("DoReport")
	end
end


function addon:OnEnable()
	addon:Print("RBS: Activated");
	addon:RegisterEvent("RAID_ROSTER_UPDATE", "JoinedPartyRaid")
	addon:RegisterEvent("PARTY_MEMBERS_CHANGED", "JoinedPartyRaid")
	addon:RegisterEvent("PLAYER_ENTERING_WORLD", "JoinedPartyRaid")
	addon:SetupFrames()
	addon:DoReport()
	if raid.isparty or raid.israid then
		addon:ShowReportFrame()
	end
end

function addon:OnDisable()
	addon:UnregisterEvent("RAID_ROSTER_UPDATE")
	addon:UnregisterEvent("PARTY_ROSTER_UPDATE")
	addon:UnregisterEvent("PLAYER_ENTERING_WORLD")
	addon:CancelScheduledEvent("DoReport")
end

function addon:Tooltip(self, title, list)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
	GameTooltip:SetText(title,1,1,1,1,1)
	if list then
		GameTooltip:AddLine(table.concat(list, ", "),nil,nil,nil,nil,1)
	end
	GameTooltip:Show()
end
